#!/usr/bin/env python
"""Program zloadrace.py demonstrates concurrency bug in ZODB Connection.setstate()
that leads to XXX

XXX

XXX load vs invalidation race is there on ZODB4 and ZODB3, but on ZODB5 there is
another open vs invalidation race.
"""

from ZODB import DB
import transaction
from persistent import Persistent

from wendelin.lib.testing import TestDB_ZEO, TestDB_NEO

from golang import func, defer
from golang import sync, context


# PInt is persistent integer.
class PInt(Persistent):
    def __init__(self, i):
        self.i = i

@func
def main():
    tdb = TestDB_ZEO('<zeo>')
    #tdb = TestDB_NEO('<neo>')

    tdb.setup()
    # XXX defer(tdb.teardown)

    # two ZODB client storage connections to the same server
    zstor1 = tdb.getZODBStorage()   ; defer(zstor1.close)
    zstor2 = tdb.getZODBStorage()   ; defer(zstor2.close)

    db1 = DB(zstor1)
    db2 = DB(zstor2)


    # XXX doc
    def init():
        transaction.begin()
        zconn = db1.open()

        root = zconn.root()
        root['obj1'] = PInt(0)
        root['obj2'] = PInt(0)

        transaction.commit()
        zconn.close()


    def C1(ctx, N):
        def c1():
            transaction.begin()
            zconn = db1.open()

            root = zconn.root()
            obj1 = root['obj1']
            obj2 = root['obj2']

            # obj1 - reload it from zstor
            # obj2 - get it from zconn cache
            for i in range(10*N):
                obj1._p_invalidate()

                # both objects must have the same values
                i1 = obj1.i
                i2 = obj2.i
                if i1 != i2:
                    raise AssertionError("C1: obj1.i (%d)  !=  obj2.i (%d)" % (i1, i2))

            transaction.abort()
            zconn.close()

        for i in range(N):
            print('C1.%d' % i)
            c1()


    def C2(ctx, N):
        def c2():
            transaction.begin()
            zconn = db2.open()

            root = zconn.root()
            obj1 = root['obj1']
            obj2 = root['obj2']
            obj1.i += 1
            obj2.i += 1
            assert obj1.i == obj2.i

            transaction.commit()
            zconn.close()

        for i in range(N):
            print('C2.%d' % i)
            c2()




    init()

    N = 100
    wg = sync.WorkGroup(context.background())
    wg.go(C1, N)
    wg.go(C2, N)
    wg.wait()


if __name__ == '__main__':
    main()
